/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

// A bkstr link looks like this:
// bkstr://www.site.com/something{from-to}/{from-to}/{from-to}.jpg

function nsBkstrProtocolHandler()
{
}

nsBkstrProtocolHandler.prototype = {
  _PHIID: Components.interfaces.nsIProtocolHandler,
  
  get scheme()          { return "bkstr"; },
  get protocolFlags()   { return this._PHIID.URI_NORELATIVE | this._PHIID.URI_NOAUTH },
  get defaultPort()     { return 0; },
  
  allowPort: function (aPort, aScheme)
  {
    return false;
  },
  
  newURI: function (aSpec, aOriginalCharset, aBaseURI)
  {
    const nsIStandardURL = Components.interfaces.nsIStandardURL;
    var uri = Components.classes["@mozilla.org/network/standard-url;1"].createInstance(nsIStandardURL);
    uri.init(nsIStandardURL.URLTYPE_STANDARD, 6667, aSpec, aOriginalCharset, aBaseURI);

    return uri.QueryInterface(Components.interfaces.nsIURI);
  },
  
  newChannel: function (aURI)
  {
    var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    if (!ioService.allowPort(aURI.port, aURI.scheme))
      throw Components.results.NS_ERROR_FAILURE;
    
    return new nsDummyChannel(aURI);
  },  

  QueryInterface: function (aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIProtocolHandler) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    
    return this;        
  }
};

var nsBkstrProtocolHandlerFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    return (new nsBkstrProtocolHandler()).QueryInterface(aIID);
  }
};

///////////////////////////////////////////////////////////////////////////////
// Dummy Channel used by nsBkstrProtocolHandler
function nsDummyChannel(aURI)
{
  this.URI = aURI;
  this.originalURI = aURI;
}

nsDummyChannel.prototype = {
  /////////////////////////////////////////////////////////////////////////////
  // nsISupports
  QueryInterface: function (aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIChannel) && 
        !aIID.equals(Components.interfaces.nsIRequest) && 
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },
  
  /////////////////////////////////////////////////////////////////////////////
  // nsIChannel
  loadAttributes:         null,
  contentType:            "x-application-bkstr",
  contentLength:          0,
  owner:                  null,
  loadGroup:              null,
  notificationCallbacks:  null,
  securityInfo:           null,
  
  open: function ()
  {
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  },
  
  asyncOpen: function (aObserver, aContext)
  {
    aObserver.onStartRequest(this, aContext);
  },
  
  asyncRead: function (aListener, aContext)
  {
    return aListener.onStartRequest(this, aContext);
  },

  /////////////////////////////////////////////////////////////////////////////
  // nsIRequest
  isPending: function () 
  { 
    return true;
  },
  
  _status: Components.results.NS_OK,
  cancel: function (aStatus) 
  { 
    this._status = aStatus; 
  },
  
  suspend: function ()
  {
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  },
  
  resume: function ()
  {
    throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
  }
};

///////////////////////////////////////////////////////////////////////////////
// Content handler for "x-application-bkstr"
function nsBkstrContentHandler()
{
}

nsBkstrContentHandler.prototype = {
  kDefaultDir          : "bengoodger.com.tabdownloader.defaultdir",
  kUseDefaultDir       : "bengoodger.com.tabdownloader.usedefaultdir",
  kFlattenHierarchy    : "bengoodger.com.tabdownloader.bkstr.flattenHierarchy",
  
  _pref : Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch),
  
  QueryInterface: function (aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIContentHandler))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },
  
  _padZeros: function (aIndex, aStartValue)
  {
    // the start value is always an indicator of how much padding should be used. 
    var padded = "";
    var indexStr = aIndex.toString();
    if (indexStr.length < aStartValue.length) {
      for (var i = 0; i < (aStartValue.length - indexStr.length); ++i)
        padded += "0";
    }
    return padded += indexStr;
  },

  _getSaveFolder: function ()
  {
    var useDefaultDir = true;
    var initialDir = null;
    try {
      // Ordering here is IMPORTANT. If the useDefaultDir pref is not set, and the ordering were 
      // reversed, an exception would be thrown and the default dir would never be retrieved, regardless
      // of whether or not it is set.
      initialDir = this._pref.getComplexValue(this.kDefaultDir, Components.interfaces.nsILocalFile);
      useDefaultDir = this._pref.getBoolPref(this.kUseDefaultDir);
    }
    catch (e) {
    }
    
    var theFile = null;
    
    if (!useDefaultDir || !initialDir) {
      var fp = Components.classes["@mozilla.org/filepicker;1"].createInstance(Components.interfaces.nsIFilePicker);
      
      fp.init(null, "Choose Save Folder", Components.interfaces.nsIFilePicker.modeGetFolder);
      fp.appendFilters(Components.interfaces.nsIFilePicker.filterAll);
      if (initialDir)
        fp.displayDirectory = initialDir;
      if (fp.show() == Components.interfaces.nsIFilePicker.returnOK) {
        this._pref.setComplexValue(this.kDefaultDir, Components.interfaces.nsILocalFile, fp.file);
        theFile = fp.file.clone();
      }  
      else 
        return initialDir;
    }
    else
      theFile = initialDir;  
    
    theFile.append(this._getDateString());  
    if (!theFile.exists()) 
      theFile.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0755);

    return theFile;
  },

  _getDateString: function ()
  {
    var date = new Date();
    var dateString = "";
    dateString += date.getFullYear() + "-";
    dateString += this._padZeros2(date.getMonth(), true) + "-";
    dateString += this._padZeros2(date.getDate(), false) + " ";
    dateString += this._padZeros2(date.getHours(), false) + "-";
    dateString += this._padZeros2(date.getMinutes(), false) + "-";
    dateString += this._padZeros2(date.getSeconds(), false);
    return dateString;
  },

  _padZeros2: function (aNumber, aShouldIncrement)
  {
    if (aShouldIncrement)
      ++aNumber;
    if (aNumber < 10) 
      return "0" + aNumber;
    return aNumber;
  },

  handleContent: function (aContentType, aCommand, aWindowTarget, aRequest)
  {
    var spec = unescape(aRequest.QueryInterface(Components.interfaces.nsIChannel).URI.spec);

    // 1) Parse the {from-to} sections out of the URL:
    var re = /\{(\d+)-(\d+)\}/g;
    var matches = spec.match(re);
    
    var from, to;
    var urls = [];
    for (var i = 0; i < matches.length; ++i) {
      var parts = matches[i].split("-");
      
      from = parts[0].substr(1, parts[0].length);
      to = parts[1].substr(0, parts[1].length - 1);
      
      var j, k;
      var suburls = [];
      
      if (urls.length == 0) {
        for (j = parseInt(from); j <= parseInt(to); ++j) {
          var match = matches[i];
          match = match.replace(/\{/, "\{");
          match = match.replace(/\}/, "\}");
          var subst = this._padZeros(j, from);
          suburls.push(spec.replace(new RegExp(match), subst));
        }
      }
      else {
        var lastSet = urls[urls.length - 1];
        for (k = 0; k < lastSet.length; ++k) {
          for (j = parseInt(from); j <= parseInt(to); ++j) {
            var match = matches[i];
            match = match.replace(/\{/, "\{");
            match = match.replace(/\}/, "\}");
            
            var subst = this._padZeros(j, from);
            suburls.push(lastSet[k].replace(new RegExp(match), subst));
          }
        }
      }
      urls.push(suburls);
    }

    // the last set contains the final URLs - save them. 
    var lastSet = urls[urls.length-1];
    var flatten = this._pref.getBoolPref(this.kFlattenHierarchy);
    var saveFolder = this._getSaveFolder();
    for (i = 0; i < lastSet.length; ++i) {
      var uri = Components.classes["@mozilla.org/network/standard-url;1"]
                          .createInstance(Components.interfaces.nsIURI);
      uri.spec = lastSet[i];
      uri.scheme = "http";
      
      // There are two modes of saving here - one is to create a mirror hierarchy on 
      // disk of the remote folder structure, the other is to flatten all the file names.
      var wbp = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"]
                          .createInstance(Components.interfaces.nsIWebBrowserPersist); 
      var file = saveFolder.clone();
      var url = uri.QueryInterface(Components.interfaces.nsIURL);
      if (flatten) {
        var fileName = this._padZeros(i+1, from) + "." + url.fileExtension;
        file.append(fileName);
      }
      else {
        var base = uri.host;
        file.append(uri.host);
        if (!file.exists()) 
          file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0755);
        
        var remainder = uri.path;
        var parts = remainder.split("/");
        for (j = 0; j < parts.length; ++j) {
          file.append(unescape(parts[j]));
          if (parts[j].length && j != parts.length - 1) {
            if (!file.exists())
              file.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0755);
          }
        }
      }

      wbp.saveURI(uri, null, null, null, null, file);
    }
  }
}

var nsBkstrContentHandlerFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    if (!aIID.equals(Components.interfaces.nsIContentHandler) && 
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_INVALID_ARG;
    
    return new nsBkstrContentHandler();
  }
}

///////////////////////////////////////////////////////////////////////////////
// Module
var nsBkstrProtocolModule = {
  _firstTime: true,
  
  registerSelf: function (aComponentManager, aFileSpec, aLocation, aType)
  {
    if (this._firstTime) {
      this._firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aComponentManager = aComponentManager.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aComponentManager.registerFactoryLocation(this._protocolHandlerCID, 
                                              "Bkstr Protocol Handler",
                                              this._protocolHandlerProgID,
                                              aFileSpec,
                                              aLocation,
                                              aType);
    aComponentManager.registerFactoryLocation(this._contentHandlerCID, 
                                              "Bkstr Content Handler",
                                              this._contentHandlerProgID,
                                              aFileSpec,
                                              aLocation,
                                              aType);
  },
  
  getClassObject: function (aComponentManager, aCID, aIID)
  {
    if (aCID.equals(this._protocolHandlerCID))
      return nsBkstrProtocolHandlerFactory;
  
    if (aCID.equals(this._contentHandlerCID))
      return nsBkstrContentHandlerFactory;

    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },
  

  _protocolHandlerCID: Components.ID("{3ce86ccb-d5e3-4ed4-8efd-87d0c4b50de8}"),
  _protocolHandlerProgID: "@mozilla.org/network/protocol;1?name=bkstr",
  _contentHandlerCID: Components.ID("{e9b0368d-ffa9-4cb4-96af-a438c3c1aac9}"),
  _contentHandlerProgID: "@mozilla.org/uriloader/content-handler;1?type=x-application-bkstr",

  canUnload: function (aComponentManager)
  {
    return true;
  }
}

function NSGetModule(aComponentManager, aFileSpec)
{
  return nsBkstrProtocolModule;
}

